import itertools
from .system import System
from ..forcefield import *
from ..topology import *
from ..trajectory import *
from .. import logger


class GromacsExporter:
    '''
    GromacsExporter export a :class:`System` to input files for Gromacs.

    The following potential functions are currently supported:

    * :class:`~mstk.forcefield.LJ126Term`
    * :class:`~mstk.forcefield.HarmonicBondTerm`
    * :class:`~mstk.forcefield.MorseBondTerm`
    * :class:`~mstk.forcefield.HarmonicAngleTerm`
    * :class:`~mstk.forcefield.HarmonicCosineAngleTerm`
    * :class:`~mstk.forcefield.LinearAngleTerm`
    * :class:`~mstk.forcefield.OplsDihedralTerm`
    * :class:`~mstk.forcefield.PeriodicDihedralTerm`
    * :class:`~mstk.forcefield.OplsImproperTerm`
    * :class:`~mstk.forcefield.HarmonicImproperTerm`
    * :class:`~mstk.forcefield.DrudePolarTerm`

    Note that the Drude polarization in GROMACS have never been thoroughly tested. Use it with care.

    In order to use GROMACS for trajectory analysis, the following potential functions can also be exported.
    But the exported files are only used to generate tpr file, which is required by most gmx analysis commands.
    Do NOT use the exported files for production simulations.

    * :class:`~mstk.forcefield.MieTerm` will be exported in the LJ-12-6 form
    * :class:`~mstk.forcefield.MorseVdwTerm` will be exported in the LJ-12-6 form
    * :class:`~mstk.forcefield.QuarticBondTerm` will be exported in the harmonic form
    * :class:`~mstk.forcefield.QuarticAngleTerm` will be exported in the harmonic form
    * :class:`~mstk.forcefield.SDKAngleTerm` will be exported in the harmonic form
    '''

    def __init__(self):
        pass

    @staticmethod
    def export(system, gro_out, top_out, mdp_out, **kwargs):
        '''
        Generate input files for Gromacs from a system

        # TODO Export virtual sites other than TIP4P

        Parameters
        ----------
        system : System
        gro_out : str or None
        top_out : str or None
        mdp_out : str or None
        '''
        if not system.use_pbc:
            raise Exception('PBC required for exporting GROMACS')

        supported_terms = {LJ126Term, MieTerm, MorseVdwTerm,
                           HarmonicBondTerm, QuarticBondTerm, MorseBondTerm,
                           HarmonicAngleTerm, QuarticAngleTerm, HarmonicCosineAngleTerm, SDKAngleTerm, LinearAngleTerm,
                           OplsDihedralTerm, PeriodicDihedralTerm,
                           OplsImproperTerm, HarmonicImproperTerm,
                           DrudePolarTerm}
        unsupported = system.ff_classes - supported_terms
        if unsupported != set():
            raise Exception('Unsupported FF terms: %s' % (', '.join(map(lambda x: x.__name__, unsupported))))

        if MieTerm in system.ff_classes:
            logger.warning('MieTerm not supported by GROMACS. Exported in LJ-12-6 form')
        if MorseVdwTerm in system.ff_classes:
            logger.warning('MorseVdwTerm not supported by GROMACS. Exported in LJ-12-6 form')
        if QuarticBondTerm in system.ff_classes:
            logger.warning('QuarticBondTerm not supported by GROMACS. Exported in harmonic form')
        if QuarticAngleTerm in system.ff_classes:
            logger.warning('QuarticAngleTerm not supported by GROMACS. Exported in harmonic form')
        if SDKAngleTerm in system.ff_classes:
            logger.warning('SDKAngleTerm not supported by GROMACS. Exported in harmonic form')
        if DrudePolarTerm in system.ff_classes:
            logger.warning('DrudePolarTerm not well tested in GROMACS. Use it with caution')

        if system.vsite_types - {TIP4PSite} != set():
            raise Exception('Virtual sites other than TIP4PSite haven\'t been implemented')

        if gro_out is not None:
            GromacsExporter._export_gro(system, gro_out)
        if top_out is not None:
            GromacsExporter._export_top(system, top_out)
        if mdp_out is not None:
            GromacsExporter._export_mdp(system, mdp_out)

    @staticmethod
    def _export_gro(system: System, gro_out='conf.gro'):
        frame = Frame(system._topology.n_atom)
        frame.cell = system._topology.cell
        frame.positions = system._topology.positions
        gro = Trajectory(gro_out, 'w')
        gro.write_frame(frame, system._topology)
        gro.close()

    @staticmethod
    def _export_top(system: System, top_out='topol.top'):
        mols_unique = system._topology.get_unique_molecules(deepcopy=False)

        string = '; GROMACS topol file created by mstk\n'

        string += '\n[ defaults ]\n'
        string += '; nbfunc  comb-rule  gen-pairs  fudgeLJ  fudgeQQ\n'
        _comb = 2 if system.ff.lj_mixing_rule == ForceField.LJ_MIXING_LB else 3
        string += '%6i %6i %12s %10.4f %10.4f\n' % (
            1, _comb, 'yes', system.ff.scale_14_vdw, system.ff.scale_14_coulomb)

        string += '\n[ atomtypes ]\n'
        string += ';     name       mass     charge      ptype      sigma      epsilon\n'
        drude_types = set()
        if DrudePolarTerm in system.ff_classes:
            for atom in system.topology.atoms:
                if atom.is_drude:
                    drude_types.add(atom.type)
        for atype in system.ff.atom_types.values():
            ptype = 'S' if atype.name in drude_types else 'A'
            vdw = system.ff.get_vdw_term(atype, atype)
            if vdw.__class__ in (LJ126Term, MieTerm):
                string += '%10s %10.4f %12.6f %6s %12.6f %12.6f  ; %s\n' % (
                    atype.name, 0.0, 0.0, ptype, vdw.sigma, vdw.epsilon, ' '.join(vdw.comments))
            elif vdw.__class__ == MorseVdwTerm:
                string += '%10s %10.4f %12.6f %6s %12.6f %12.6f  ; %s\n' % (
                    atype.name, 0.0, 0.0, ptype, vdw.r0, vdw.depth, ' '.join(vdw.comments))
            else:
                raise Exception('Unsupported vdW term')

        string += '\n[ nonbond_params ]\n'
        string += ';       i        j        func      sigma      epsilon\n'
        for at1, at2 in itertools.combinations(system.ff.atom_types.values(), 2):
            try:
                vdw = system.ff.get_vdw_term(at1, at2, mixing=False)
            except:
                continue
            if vdw.__class__ in (LJ126Term, MieTerm):
                string += '%10s %10s %6i %12.6f %12.6f  ; %s\n' % (
                    at1.name, at2.name, 1, vdw.sigma, vdw.epsilon, ' '.join(vdw.comments))
            elif vdw.__class__ == MorseVdwTerm:
                string += '%10s %10s %6i %12.6f %12.6f  ; %s\n' % (
                    at1.name, at2.name, 1, vdw.r0, vdw.depth, ' '.join(vdw.comments))
            else:
                raise Exception('Unsupported vdW term')

        for i, mol in enumerate(mols_unique.keys()):
            mol: Molecule
            string += '\n[ moleculetype ]\n'
            string += '; Name            nrexcl\n'
            string += f'{mol.name}_{i + 1}     3\n'

            string += '\n[ atoms ]\n'
            string += ';%5s %10s %6s %10s %6s %6s %12s %10s\n' % (
                'nr', 'type', 'resnr', 'residue', 'atom', 'cgnr', 'charge', 'mass')
            for j, atom in enumerate(mol.atoms):
                # GROMACS use SCF approach for Drude simulation. Set mass of Drude to zero
                mass = 0 if atom.is_drude else atom.mass
                if atom in system.drude_pairs:
                    mass += system.drude_pairs[atom].mass
                string += '%6i %10s %6i %10s %6s %6i %12.6f %10.4f\n' % (
                    j + 1, atom.type, atom.residue.id_in_mol + 1, atom.residue.name,
                    atom.symbol, j + 1, atom.charge, mass)

            if TIP4PSite in system.vsite_types:
                string += '\n[ virtual_sites3 ]\n'
                string += ';%5s %6s %6s %6s %6s %12s %12s\n' % ('Vsite', 'from1', 'from2', 'from3', 'funct', 'a', 'b')
                for atom in mol.atoms:
                    vsite = atom.virtual_site
                    if type(vsite) is TIP4PSite:
                        O, H1, H2 = vsite.parents
                        w_O, w_H1, w_H2 = system.get_TIP4P_linear_coeffs(atom)
                        string += '%6i %6i %6i %6i %6i %12.6f %12.6f\n' % (
                            atom.id_in_mol + 1, O.id_in_mol + 1, H1.id_in_mol + 1, H2.id_in_mol + 1, 1, w_H1, w_H2)
                    elif vsite is not None:
                        raise Exception('Virtual sites other than TIP4PSite haven\'t been implemented')

            string += '\n[ pairs ]\n'
            pairs12, pairs13, pairs14 = mol.get_12_13_14_pairs()
            for a1, a2 in pairs14:
                string += '%6i %6i %6i\n' % (a1.id_in_mol + 1, a2.id_in_mol + 1, 1)

            # 1-4 pairs concerning Drude particles
            drude_pairs = dict(mol.get_drude_pairs())
            _pairs14_drude = []
            for a1, a2 in pairs14:
                if a1 in drude_pairs:
                    _pairs14_drude.append((drude_pairs[a1], a2))
                if a2 in drude_pairs:
                    _pairs14_drude.append((a1, drude_pairs[a2]))
                if a1 in drude_pairs and a2 in drude_pairs:
                    _pairs14_drude.append((drude_pairs[a1], drude_pairs[a2]))

            # 1-4 pairs concerning virtual sites
            vsite_pairs = dict(mol.get_virtual_site_pairs())
            _pairs14_vsite = []
            for atom1, atom2 in pairs14:
                vsite1 = vsite_pairs.get(atom1)
                vsite2 = vsite_pairs.get(atom2)
                drude1 = drude_pairs.get(atom1)
                drude2 = drude_pairs.get(atom2)
                if vsite1 is not None:
                    _pairs14_vsite.append((vsite1, atom2))
                    if drude2 is not None:
                        _pairs14_vsite.append((vsite1, drude2))
                if vsite2 is not None:
                    _pairs14_vsite.append((vsite2, atom1))
                    if drude1 is not None:
                        _pairs14_vsite.append((vsite2, drude1))
                if None not in [vsite1, vsite2]:
                    _pairs14_vsite.append((vsite1, vsite2))

            for a1, a2 in _pairs14_drude + _pairs14_vsite:
                string += '%6i %6i %6i\n' % (a1.id_in_mol + 1, a2.id_in_mol + 1, 1)

            ### Drude polarization #############################################
            if DrudePolarTerm in system.ff_classes:
                string += '\n[ polarization ]\n'
                string += ';  ai    aj   func    alpha     delta     khyp\n'
                for parent, drude in drude_pairs.items():
                    string += '%5d %5d   %4d  %9.6f  %7.4f  %9.4e\n' % (
                        parent.id_in_mol + 1, drude.id_in_mol + 1,
                        2, parent.alpha, 0.02, 16.768e8)

                string += '\n[ thole_polarization ]\n'
                string += ';  ai    aj    ak    al   func   thole   alpha1   alpha2\n'
                for a1, a2 in pairs12 + pairs13:
                    if a1 in drude_pairs and a2 in drude_pairs:
                        string += '%5i %5i %5i %5i   %4i   %9.5f  %9.6f  %9.6f\n' % (
                            a1.id_in_mol + 1, drude_pairs[a1].id_in_mol + 1,
                            a2.id_in_mol + 1, drude_pairs[a2].id_in_mol + 1,
                            1, (a1.thole + a2.thole) / 2, a1.alpha, a2.alpha)

            string += '\n[ constraints ]\n'
            for bond in mol.bonds:
                distance = system.constrain_bonds.get(bond)
                if distance is not None:
                    a1, a2 = bond.atom1, bond.atom2
                    string += '%6i %6i %6i %12.6f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1, 1, distance)
            for angle in mol.angles:
                distance = system.constrain_angles.get(angle)
                if distance is not None:
                    a1, a3 = angle.atom1, angle.atom3
                    string += '%6i %6i %6i %12.6f\n' % (
                        a1.id_in_mol + 1, a3.id_in_mol + 1, 2, distance)

            string += '\n[ bonds ]\n'
            for bond in mol.bonds:
                if bond.is_drude:
                    continue
                if bond in system.constrain_bonds:
                    continue
                bterm = system.bond_terms[bond]
                if bterm.__class__ == HarmonicBondTerm:
                    a1, a2 = bond.atom1, bond.atom2
                    string += '%6i %6i %6i %12.6f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1,
                        1, bterm.length, bterm.k * 2)
                elif bterm.__class__ == QuarticBondTerm:
                    a1, a2 = bond.atom1, bond.atom2
                    string += '%6i %6i %6i %12.6f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1,
                        1, bterm.length, bterm.k2 * 2)
                elif bterm.__class__ == MorseBondTerm:
                    a1, a2 = bond.atom1, bond.atom2
                    string += '%6i %6i %6i %12.6f %12.4f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1,
                        3, bterm.length, bterm.depth, (bterm.k / bterm.depth) ** 0.5)
                else:
                    raise Exception('Unsupported bond term')

            # Exclusions concerning Drude particles
            _exclusions_drude = {}  # {drude1: [atom2, drude2, atom3, drude3, ...]}
            for a1, a2 in pairs12 + pairs13 + pairs14:
                if a1 in drude_pairs:
                    d1 = drude_pairs[a1]
                    if d1 not in _exclusions_drude:
                        _exclusions_drude[d1] = []
                    _exclusions_drude[d1].append(a2)
                if a2 in drude_pairs:
                    d2 = drude_pairs[a2]
                    if d2 not in _exclusions_drude:
                        _exclusions_drude[d2] = []
                    _exclusions_drude[d2].append(a1)
                if a1 in drude_pairs and a2 in drude_pairs:
                    _exclusions_drude[d1].append(d2)

            # Exclusions concerning virtual sites
            _exclusions_vsite = {}  # {avsite1: [atom1, drude1, atom2, drude2, avsite2, ...]}
            for atom, vsite in vsite_pairs.items():
                _exclusions_vsite[vsite] = [atom]
                drude = drude_pairs.get(atom)
                if drude is not None:
                    _exclusions_vsite[vsite].append(drude)
            for atom1, atom2 in pairs12 + pairs13:
                vsite1 = vsite_pairs.get(atom1)
                vsite2 = vsite_pairs.get(atom2)
                drude1 = drude_pairs.get(atom1)
                drude2 = drude_pairs.get(atom2)
                if vsite1 is not None:
                    _exclusions_vsite[vsite1].append(atom2)
                    if drude2 is not None:
                        _exclusions_vsite[vsite1].append(drude2)
                if vsite2 is not None:
                    _exclusions_vsite[vsite2].append(atom1)
                    if drude1 is not None:
                        _exclusions_vsite[vsite2].append(drude1)
                if None not in [vsite1, vsite2]:
                    _exclusions_vsite[vsite1].append(vsite2)

            _exclusions = {**_exclusions_drude, **_exclusions_vsite}
            if len(_exclusions) > 0:
                string += '\n[ exclusions ]\n'
                string += ';  ai    aj    ...\n'
                for a1, a2_list in _exclusions.items():
                    string += '%5d' % (a1.id_in_mol + 1)
                    for a2 in a2_list:
                        string += ' %5d' % (a2.id_in_mol + 1)
                    string += '\n'

            string += '\n[ angles ]\n'
            for angle in mol.angles:
                if angle in system.constrain_angles:
                    continue
                aterm = system.angle_terms[angle]
                a1, a2, a3 = angle.atom1, angle.atom2, angle.atom3
                if aterm.__class__ in (HarmonicAngleTerm, SDKAngleTerm):
                    string += '%6i %6i %6i %6i %12.6f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1, a3.id_in_mol + 1,
                        1, aterm.theta * RAD2DEG, aterm.k * 2)
                elif aterm.__class__ is QuarticAngleTerm:
                    string += '%6i %6i %6i %6i %12.6f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1, a3.id_in_mol + 1,
                        1, aterm.theta * RAD2DEG, aterm.k2 * 2)
                elif aterm.__class__ is HarmonicCosineAngleTerm:
                    string += '%6i %6i %6i %6i %12.6f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1, a3.id_in_mol + 1,
                        2, aterm.theta * RAD2DEG, aterm.k * 2 / math.sin(aterm.theta) ** 2)
                elif aterm.__class__ is LinearAngleTerm:
                    bond12, bond23 = angle.bonds
                    bterm12 = system.bond_terms[bond12]
                    bterm23 = system.bond_terms[bond23]
                    a, k_linear = aterm.calc_a_k_linear(bterm12.length, bterm23.length)
                    string += '%6i %6i %6i %6i %12.6f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1, a3.id_in_mol + 1,
                        9, a, k_linear * 2)
                else:
                    raise Exception('Unsupported angle term')

            string += '\n[ dihedrals ]\n'
            for dihedral in mol.dihedrals:
                dterm = system.dihedral_terms[dihedral]
                if dterm.__class__ == OplsDihedralTerm:
                    dterm = dterm.to_periodic_term()
                a1, a2, a3, a4 = dihedral.atom1, dihedral.atom2, dihedral.atom3, dihedral.atom4
                if dterm.__class__ == PeriodicDihedralTerm:
                    for phase in dterm.phases:
                        string += '%6i %6i %6i %6i %6i %8.2f %12.4f %4i\n' % (
                            a1.id_in_mol + 1, a2.id_in_mol + 1,
                            a3.id_in_mol + 1, a4.id_in_mol + 1,
                            9, phase.phi * RAD2DEG, phase.k, phase.n)
                    if len(dterm.phases) == 0:
                        string += '%6i %6i %6i %6i %6i %8.2f %12.4f %4i  ; no dihedral parameters\n' % (
                            a1.id_in_mol + 1, a2.id_in_mol + 1,
                            a3.id_in_mol + 1, a4.id_in_mol + 1,
                            9, 0.0, 0.0, 1)
                else:
                    raise Exception('Unsupported dihedral term')

            string += '\n[ dihedrals ]\n'
            for improper in mol.impropers:
                iterm = system.improper_terms[improper]
                a1, a2, a3, a4 = improper.atom1, improper.atom2, improper.atom3, improper.atom4
                if iterm.__class__ == OplsImproperTerm:
                    # be careful about the sequence of atoms in OPLS improper definition
                    string += '%6i %6i %6i %6i %6i %8.2f %12.4f %4i\n' % (
                        a2.id_in_mol + 1, a3.id_in_mol + 1,
                        a1.id_in_mol + 1, a4.id_in_mol + 1,
                        4, 180, iterm.k, 2)
                elif iterm.__class__ == HarmonicImproperTerm:
                    string += '%6i %6i %6i %6i %6i %8.2f %12.4f\n' % (
                        a1.id_in_mol + 1, a2.id_in_mol + 1,
                        a3.id_in_mol + 1, a4.id_in_mol + 1,
                        2, iterm.phi * RAD2DEG, iterm.k * 2)
                else:
                    raise Exception('Unsupported improper term')

        string += '\n[ system ]\n'

        string += '\n[ molecules ]\n'
        for i, (mol, count) in enumerate(mols_unique.items()):
            string += '%s_%i %6i\n' % (mol.name, i + 1, count)

        with open(top_out, 'wb') as f:
            f.write(string.encode())

    @staticmethod
    def _export_mdp(system: System, mdp_out='grompp.mdp'):
        r_cut = system.ff.vdw_cutoff
        tau_t = 0.2 if DrudePolarTerm in system.ff_classes else 1.0
        string = f'''; Created by mstk
integrator      = sd
dt              = 0.002 ; ps
nsteps          = 1000000

nstxout         = 0
nstvout         = 0
nstfout         = 0
nstxout-compressed = 1000
compressed-x-grps  = System

cutoff-scheme   = Verlet
pbc             = xyz
; rlist           = {r_cut}
coulombtype     = PME
rcoulomb        = {r_cut}
vdwtype         = Cut-off
rvdw            = {r_cut}
DispCorr        = EnerPres

tcoupl          = no; v-rescale
tc_grps         = System
tau_t           = {tau_t}
ref_t           = 300

pcoupl          = berendsen ; parrinello-rahman
pcoupltype      = isotropic
tau_p           = 0.5; 5
compressibility = 4.5e-5
ref_p           = 1.0

gen_vel         = yes
gen_temp        = 300

constraints     = h-bonds
constraint-algorithm = LINCS
'''
        with open(mdp_out, 'wb') as f:
            f.write(string.encode())



import os
from .system import System
from ..forcefield import *
from ..topology import *
from .. import logger


class LammpsExporter:
    '''
    LammpsExporter export a non-polarizable :class:`System` to input files for LAMMPS.

    LAMMPS is powerful and flexible. But the input file for LAMMPS is a mess and cannot be handled in an elegant way,
    especially when there are hybrid functional forms.
    Here, only limited types of FF (OPLS-AA and SDK-CG) will be considered.

    The following potential functions are currently supported:

    * :class:`~mstk.forcefield.LJ126Term`
    * :class:`~mstk.forcefield.MieTerm`
    * :class:`~mstk.forcefield.HarmonicBondTerm`
    * :class:`~mstk.forcefield.HarmonicAngleTerm`
    * :class:`~mstk.forcefield.SDKAngleTerm`
    * :class:`~mstk.forcefield.OplsDihedralTerm`
    * :class:`~mstk.forcefield.PeriodicDihedralTerm`
    * :class:`~mstk.forcefield.OplsImproperTerm`

    The :class:~mstk.forcefield.LinearAngleTerm can be exported, but it will be in the harmonic form.
    This is only valid when the force constant is large enough to keep the angle nearby 180 degree.

    In order to run simulation of SDK-CG model, the package CG-SDK should be included when compiling LAMMPS.
    '''

    def __init__(self):
        pass

    @staticmethod
    def export(system, data_out, in_out, **kwargs):
        '''
        Generate input files for LAMMPS from a system

        Parameters
        ----------
        system : System
        data_out : str
        in_out : str
        '''
        if not system.use_pbc:
            raise Exception('PBC required for exporting LAMMPS')

        supported_terms = {LJ126Term, MieTerm,
                           HarmonicBondTerm, MorseBondTerm,
                           HarmonicAngleTerm, SDKAngleTerm, LinearAngleTerm,
                           OplsDihedralTerm, PeriodicDihedralTerm,
                           OplsImproperTerm}
        unsupported = system.ff_classes - supported_terms
        if unsupported != set():
            raise Exception('Unsupported FF terms: %s' % (', '.join(map(lambda x: x.__name__, unsupported))))

        if system.topology.has_virtual_site:
            raise Exception('Virtual sites not supported by LAMMPS')

        if LinearAngleTerm in system.ff_classes:
            logger.warning('LinearAngleTerm not supported by LAMMPS. Exported in harmonic form')

        top = system.topology
        ff = system.ff

        # Morse bond requires hybrid bond styles
        is_morse = MorseBondTerm in system.ff_classes

        # SDK CGFF requires all pairwise parameters and hybrid angle styles
        is_sdk = MieTerm in system.ff_classes or SDKAngleTerm in system.ff_classes

        ### Assign LAMMPS types ####################################
        bond_types = list(ff.bond_terms.values())
        angle_types = list(ff.angle_terms.values())
        dihedral_types = list(ff.dihedral_terms.values())
        improper_types = list(ff.improper_terms.values())

        lmp_types: {str: Atom} = {}  # {'typ': Atom}
        lmp_types_atype: {str: AtomType} = {}  # {'typ': AtomType}

        for atom in top.atoms:
            if atom.type not in lmp_types:
                lmp_types[atom.type] = atom
                lmp_types_atype[atom.type] = ff.atom_types[atom.type]

        lmp_type_list = list(lmp_types.keys())
        lmp_symbol_list = [atom.symbol or 'UNK' for atom in lmp_types.values()]
        lmp_shake_bonds = [i + 1 for i, bterm in enumerate(bond_types) if bterm.fixed]
        lmp_shake_angles = [i + 1 for i, aterm in enumerate(angle_types) if aterm.fixed]
        #############################################################

        string = 'Created by mstk\n'

        string += '\n%i atoms\n' % top.n_atom
        string += '%i bonds\n' % top.n_bond
        string += '%i angles\n' % top.n_angle
        string += '%i dihedrals\n' % top.n_dihedral
        string += '%i impropers\n' % top.n_improper

        string += '\n%i atom types\n' % len(lmp_type_list)
        string += '%i bond types\n' % len(bond_types)
        string += '%i angle types\n' % len(angle_types)
        string += '%i dihedral types\n' % len(dihedral_types)
        string += '%i improper types\n' % len(improper_types)

        if not top.cell.is_rectangular:
            raise Exception('Triclinic box haven\'t been implemented')
        box = top.cell.get_size() * 10
        string += '\n0 %10.4f  xlo xhi\n' % box[0]
        string += '0 %10.4f  ylo yhi\n' % box[1]
        string += '0 %10.4f  zlo zhi\n' % box[2]

        string += '\nMasses\n\n'

        for i, (typ, atom) in enumerate(lmp_types.items()):
            string += '%4i %8.3f  # %8s\n' % (i + 1, atom.mass, typ)

        if len(bond_types) > 0:
            string += '\nBond Coeffs  # %s\n\n' % ('hybrid harmonic morse' if is_morse else 'harmonic')

        for i, bterm in enumerate(bond_types):
            bond_style = ''
            if is_morse:
                bond_style = 'morse' if type(bterm) is MorseBondTerm else 'harmonic'
            if type(bterm) is HarmonicBondTerm:
                string += '%4i %8s %12.6f %10.4f  # %s-%s\n' % (
                    i + 1, bond_style, bterm.k / 4.184 / 100, bterm.length * 10, bterm.type1, bterm.type2)
            elif type(bterm) is MorseBondTerm:
                alpha = (bterm.k / bterm.depth) ** 0.5
                string += '%4i %8s %12.6f %10.4f %10.4f  # %s-%s\n' % (
                    i + 1, bond_style, bterm.depth / 4.184, alpha / 10, bterm.length * 10, bterm.type1, bterm.type2)

        if len(angle_types) > 0:
            string += '\nAngle Coeffs  # %s\n\n' % ('hybrid harmonic sdk' if is_sdk else 'harmonic')

        for i, aterm in enumerate(angle_types):
            angle_style = ''
            if is_sdk:
                angle_style = 'sdk' if type(aterm) is SDKAngleTerm else 'harmonic'
            string += '%4i %8s %12.6f %10.4f  # %s-%s-%s\n' % (
                i + 1, angle_style, aterm.k / 4.184, aterm.theta * RAD2DEG, aterm.type1, aterm.type2, aterm.type3)

        if len(dihedral_types) > 0:
            string += '\nDihedral Coeffs  # opls\n\n'

        for i, dterm in enumerate(dihedral_types):
            if type(dterm) is PeriodicDihedralTerm:
                try:
                    dterm = dterm.to_opls_term()
                except:
                    raise Exception('Only OPLS type dihedral is implemented')
            k1, k2, k3, k4 = map(lambda x: x * 2 / 4.184, [dterm.k1, dterm.k2, dterm.k3, dterm.k4])
            string += '%4i %12.6f %12.6f %12.6f %12.6f  # %s-%s-%s-%s\n' % (
                i + 1, k1, k2, k3, k4,
                dterm.type1, dterm.type2, dterm.type3, dterm.type4)

        if len(improper_types) > 0:
            string += '\nImproper Coeffs  # cvff\n\n'

        for i, iterm in enumerate(improper_types):
            string += '%4i %12.6f %4i %4i  # %s-%s-%s-%s\n' % (
                i + 1, iterm.k / 4.184, -1, 2, iterm.type2, iterm.type3, iterm.type1, iterm.type4)

        string += '\nAtoms\n\n'

        for atom in top.atoms:
            typ = atom.type
            x, y, z = atom.position * 10
            string += '%8i %6i %6i %12.6f %10.4f %10.4f %10.4f  # %8s %8s\n' % (
                atom.id + 1, atom.molecule.id + 1, lmp_type_list.index(typ) + 1,
                atom.charge, x, y, z, atom.name, atom.molecule.name)

        if top.n_bond > 0:
            string += '\nBonds\n\n'

        for i, bond in enumerate(top.bonds):
            a1, a2 = bond.atom1, bond.atom2
            btype = bond_types.index(system.bond_terms[bond]) + 1
            string += '%6i %6i %6i %6i  # %s-%s\n' % (
                i + 1, btype, a1.id + 1, a2.id + 1, a1.name, a2.name)

        if top.n_angle > 0:
            string += '\nAngles\n\n'

        for i, angle in enumerate(top.angles):
            aterm = angle_types.index(system.angle_terms[angle]) + 1
            a1, a2, a3 = angle.atom1, angle.atom2, angle.atom3
            string += '%6i %6i %6i %6i %6i  # %s-%s-%s\n' % (
                i + 1, aterm, a1.id + 1, a2.id + 1, a3.id + 1, a1.name, a2.name, a3.name)

        if top.n_dihedral > 0:
            string += '\nDihedrals\n\n'

        for i, dihedral in enumerate(top.dihedrals):
            dtype = dihedral_types.index(system.dihedral_terms[dihedral]) + 1
            a1, a2, a3, a4 = dihedral.atom1, dihedral.atom2, dihedral.atom3, dihedral.atom4
            string += '%6i %6i %6i %6i %6i %6i  # %s-%s-%s-%s\n' % (
                i + 1, dtype, a1.id + 1, a2.id + 1, a3.id + 1, a4.id + 1,
                a1.name, a2.name, a3.name, a4.name)

        if top.n_improper > 0:
            string += '\nImpropers\n\n'

        for i, improper in enumerate(top.impropers):
            itype = improper_types.index(system.improper_terms[improper]) + 1
            a1, a2, a3, a4 = improper.atom1, improper.atom2, improper.atom3, improper.atom4
            string += '%6i %6i %6i %6i %6i %6i  # %s-%s-%s-%s\n' % (
                i + 1, itype, a2.id + 1, a3.id + 1, a1.id + 1, a4.id + 1,
                a2.name, a3.name, a1.name, a4.name)

        with open(data_out, 'wb') as f:
            f.write(string.encode())

        ### LAMMPS input script #################################################
        cmd_bond = 'hybrid harmonic morse' if is_morse else 'harmonic'
        cmd_angle = 'hybrid harmonic sdk' if is_sdk else 'harmonic'
        cmd_pair_base = 'lj/sdk' if is_sdk else 'lj/cut'
        cmd_pair_suffix = '/coul/long' if system.charged else ''
        cmd_mix = 'arithmetic' if system.ff.lj_mixing_rule == ForceField.LJ_MIXING_LB else 'geometric'
        cmd_tail = 'yes' if ff.vdw_long_range == ForceField.VDW_LONGRANGE_CORRECT else 'no'
        cmd_shift = 'yes' if ff.vdw_long_range == ForceField.VDW_LONGRANGE_SHIFT else 'no'
        cmd_comment_kspace = '# ' if not system.charged else ''

        cmd_paras = ''
        for i, (typ_i, atom) in enumerate(lmp_types.items()):
            for j, (typ_j, atom) in enumerate(lmp_types.items()):
                if j < i:
                    continue
                atype1 = lmp_types_atype[typ_i]
                atype2 = lmp_types_atype[typ_j]
                try:
                    # LAMMPS CG-SDK does not allow lj mixing. have to write it explicitly
                    vdw = ff.get_vdw_term(atype1, atype2, mixing=is_sdk)
                except:
                    continue

                if is_sdk:
                    if type(vdw) is MieTerm:
                        if vdw.is_sdk:
                            lj_style = f'lj{int(vdw.repulsion)}_{int(vdw.attraction)}'
                        else:
                            raise Exception('MieTerm other than SDK type are not implemented')
                    else:
                        lj_style = 'lj12_6'
                    cmd_paras += 'pair_coeff %4i %4i %8s %9.5f %8.4f  # %8s %8s %s\n' % (
                        i + 1, j + 1, lj_style, vdw.epsilon / 4.184, vdw.sigma * 10,
                        lmp_type_list[i], lmp_type_list[j], ','.join(vdw.comments))
                else:
                    cmd_paras += 'pair_coeff %4i %4i %9.5f %8.4f  # %8s %8s %s\n' % (
                        i + 1, j + 1, vdw.epsilon / 4.184, vdw.sigma * 10,
                        lmp_type_list[i], lmp_type_list[j], ','.join(vdw.comments))

        cmd_shake = 'fix SHAKE all shake 0.0001 20 0 b '
        if lmp_shake_bonds:
            cmd_shake += ' '.join(map(str, lmp_shake_bonds))
            if lmp_shake_angles:
                cmd_shake += ' a ' + ' '.join(map(str, lmp_shake_angles))
        else:
            cmd_shake = '# ' + cmd_shake
        timestep = 10.0 if is_sdk else 1.0

        string = f'''# created by mstk
units real
boundary p p p
atom_style full
bond_style {cmd_bond}
angle_style {cmd_angle}
dihedral_style opls
improper_style cvff
special_bonds lj 0 0 {ff.scale_14_vdw} coul 0 0 {ff.scale_14_coulomb}

pair_style {cmd_pair_base}{cmd_pair_suffix} {ff.vdw_cutoff * 10}
pair_modify mix {cmd_mix} tail {cmd_tail} shift {cmd_shift}
{cmd_comment_kspace}kspace_style pppm 1.0e-4

read_data {os.path.basename(data_out)}

{cmd_paras}

variable T equal 300
variable P equal 1
variable elec equal ecoul+elong

# thermo_style custom step press pe ebond eangle edihed eimp evdwl v_elec
# thermo 10
# minimize 1.0e-4 1.0e-6 200 1000
# reset_timestep 0

{cmd_shake}
fix ICECUBE all momentum 100 linear 1 1 1

velocity all create $T 12345
timestep {timestep}

fix NPT all npt temp $T $T {timestep * 100} iso $P $P {timestep * 500} fixedpoint 0 0 0

thermo_style custom step temp press pe emol evdwl v_elec density
thermo_modify flush yes
thermo 1000

variable slog equal logfreq(10,9,10)
dump TRJ all custom 10 dump.lammpstrj id mol type element q xu yu zu
dump_modify TRJ sort id element {' '.join(lmp_symbol_list)} every v_slog first yes
dump DCD all dcd 10000 dump.dcd
dump_modify DCD unwrap yes

restart 1000000 rst_*
run 1000000
'''

        with open(in_out, 'wb') as f:
            f.write(string.encode())



#!/usr/bin/env python3

import sys
import argparse
import math
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from mstk.topology import Topology
from mstk.trajectory import Trajectory
from mstk.utils import print_data_to_file
from mstk.topology.geometry import periodic_distance
from mstk import logger

np.seterr(all='raise')
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 15})


def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-p', '--top', type=str, required=True, help='topology file')
    parser.add_argument('-c', '--conf', nargs='+', type=str, required=True, help='trajectory files')
    parser.add_argument('-o', '--output', default='rdf', type=str, help='Basename for output files')
    parser.add_argument('--a1', type=str, nargs='+', help='type of central atoms')
    parser.add_argument('--a2', type=str, nargs='+', help='type of coordinate atoms')
    parser.add_argument('--m1', type=str, help='name of central molecules')
    parser.add_argument('--m2', type=str, help='name of coordinate molecules')
    parser.add_argument('--type', default='all', type=str, choices=['all', 'inter', 'intra'],
                        help='calculate RDF between all pairs, inter-molecular pairs, or intra-molecular pairs')
    parser.add_argument('-b', '--begin', default=0, type=int,
                        help='first frame to analyze. Index starts from 0')
    parser.add_argument('-e', '--end', default=-1, type=int,
                        help='last frame (not included) to analyze. Index starts from 0. '
                             '-1 means until last frames (included)')
    parser.add_argument('--skip', default=1, type=int, help='read every N frames')
    parser.add_argument('--maxr', default=1.0, type=float,
                        help='max distance (nm) for calculation of distribution')
    parser.add_argument('--dr', default=0.01, type=float,
                        help='bin size (nm) for calculation of distribution')
    parser.add_argument('--ignore', nargs='+', default=[], type=str,
                        help='ignore these molecules in topology in case topology and trajectory do not match')

    return parser.parse_args()


def main():
    args = parse_args()
    top = Topology.open(args.top)
    if args.ignore:
        molecules = [mol for mol in top.molecules if mol.name not in args.ignore]
        top.update_molecules(molecules)
    logger.info(top)

    trj = Trajectory.open(args.conf)
    logger.info(trj)

    if top.n_atom != trj.n_atom:
        raise Exception('Number of atoms in topology and trajectory files do not match')

    if args.m1:
        group1 = [atom for mol in top.molecules if mol.name == args.m1 for atom in mol.atoms if atom.type in args.a1]
    else:
        group1 = [atom for atom in top.atoms if atom.type in args.a1]

    if args.m2:
        group2 = [atom for mol in top.molecules if mol.name == args.m2 for atom in mol.atoms if atom.type in args.a2]
    else:
        group2 = [atom for atom in top.atoms if atom.type in args.a2]

    if args.end > trj.n_frame or args.end == -1:
        args.end = trj.n_frame

    dr = args.dr
    n_bin = math.ceil(args.maxr / dr) + 1
    edges = np.array([dr * (i - 0.5) for i in range(n_bin + 1)], dtype=float)
    r_array = (edges[1:] + edges[:-1]) / 2
    rdf_array = np.zeros(n_bin, dtype=float)

    n_frame = 0
    for i in range(args.begin, args.end, args.skip):
        n_frame += 1
        frame = trj.read_frame(i)
        sys.stdout.write('\r    frame %i' % i)

        vol = frame.cell.volume
        box = frame.cell.get_size()
        density_pair = len(group1) * len(group2) / vol
        density_array = np.zeros(len(r_array), dtype=float)
        for atom1 in group1:
            pos1 = frame.positions[atom1.id]
            for atom2 in group2:
                if args.type == 'inter' and atom1.molecule is atom2.molecule:
                    continue
                if args.type == 'intra' and atom1.molecule is not atom2.molecule:
                    continue
                pos2 = frame.positions[atom2.id]
                distance = periodic_distance(pos1, pos2, box, args.maxr + dr / 2)
                if distance is not None and distance < args.maxr + dr / 2:
                    idx_r = int((distance - edges[0]) / dr)
                    density_array[idx_r] += 1

        rdf_array[1:] += density_array[1:] / (4 * np.pi * r_array[1:] ** 2 * dr) / density_pair

    rdf_array = rdf_array / n_frame

    name_column_dict = {'r'  : r_array,
                        'rdf': rdf_array}
    print_data_to_file(name_column_dict, f'{args.output}.txt')

    fig, ax = plt.subplots()
    ax.set(xlabel='r (nm)', ylabel='RDF')
    ax.plot(r_array, rdf_array)
    # ax.legend()
    fig.tight_layout()
    fig.savefig(f'{args.output}.png')


if __name__ == '__main__':
    main()



import datetime
import os
import time
from dataclasses import dataclass
from mstk import logger
from mstk.misc import DocstringMeta
from .pbsjob import PbsJob


@dataclass
class JobParameter:
    '''
    The parameters for submitting a job

    Attributes
    ----------
    queue : str
        The queue to submit the job
    n_proc : int
        The number of CPU cores a job can use
    n_gpu : int
        The number of GPU cards a job can use
    n_node : int
        The number of nodes a job can use
    exclude : str
        The nodes to be excluded
    max_running_hour: int
        The wall time limit for a job in hours
    env_cmd : str
        The commands for setting up the environment before running real calculations

    '''
    queue: str = 'default'
    n_proc: int = 1
    n_gpu: int = 0
    n_node: int = 0
    exclude: str = ''
    max_running_hour: int = 24
    env_cmd: str = ''


class Scheduler(metaclass=DocstringMeta):
    '''
    Base class for job schedulers.

    Scheduler should not be constructed directly. Use its subclasses instead.

    Attributes
    ----------
    username : str
        The current user
    sh : str
        The default name of the job script
    job_parameter : JobParameter
        The default parameters for submitting a job
    cached_jobs_expire : int
        The lifetime of cached jobs in seconds.
    '''

    #: Whether this is a remote job scheduler
    is_remote = False

    def __init__(self):
        if os.name == 'nt':
            import win32api
            self.username = win32api.GetUserName()
        else:
            import pwd
            self.username = pwd.getpwuid(os.getuid()).pw_name

        self.sh = '_job.sh'
        self.job_parameter = JobParameter()
        self.cached_jobs_expire = 60  # seconds
        self._cached_jobs = []
        self._cache_last_update = None

    def is_working(self):
        '''
        Whether this job scheduler is running normally.

        Returns
        -------
        is : bool
        '''
        raise NotImplementedError('This method should be implemented by subclasses')

    def generate_sh(self, commands, name, params, workdir=None, sh=None):
        '''
        Generate a shell script for commands to be executed by the job scheduler on compute nodes.

        Parameters
        ----------
        commands : str
            List of commands to be executed by the job scheduler on compute node step by step.
        name : str
            The name of the job to be submitted.
        params : JobParameter
            The parameters for the job.
        workdir : str, Optional
            The working directory.
        sh : str, Optional
            The name (path) of the shell script being written.
            If not set, will use the default :attr:`sh`.
        '''
        raise NotImplementedError('This method should be implemented by subclasses')

    def upload(self, **kwargs) -> bool:
        '''
        Upload the simulation files to target folder.

        This method should be implemented by subclasses that is remote scheduler,
        which is determined by the attribute :attr:`is_remote`.
        If it's not a remote job scheduler, will simply return True.

        Returns
        -------
        successful : bool
            Whether the upload is successful
        '''
        pass

    def download(self, **kwargs) -> bool:
        '''
        Download the simulation files to target folder.

        This method should be implemented by subclasses that is remote scheduler,
        which is determined by the attribute :attr:`is_remote`.
        If it's not a remote job scheduler, will simply return True.

        Returns
        -------
        successful : bool
            Whether the download is successful
        '''
        pass

    def submit(self, sh=None, **kwargs):
        '''
        Submit a job script to scheduler.

        Parameters
        ----------
        sh : str, optional
            The file name of the job script.
            If not set, will use the default :attr:`sh`.

        Returns
        -------
        id : int
            Job ID. -1 means failed
        '''
        raise NotImplementedError('This method should be implemented by subclasses')

    def get_job_from_name(self, name):
        '''
        Get the job with specified name.

        If such a job can not be found, None will be returned.
        If several job have same name, the most recently submitted one will be returned.

        Parameters
        ----------
        name : str

        Returns
        -------
        job : PbsJob or None
        '''
        # if several job have same name, return the one with the largest id (most recent job)
        for job in sorted(self.get_jobs(), key=lambda x: x.id, reverse=True):
            if job.name == name:
                return job
        return None

    def is_running(self, name):
        '''
        Check whether a job is pending or running (not killed or finished or failed).

        Parameters
        ----------
        name : str

        Returns
        -------
        is : bool
        '''
        job = self.get_job_from_name(name)
        if job is None:
            return False
        return job.state in [PbsJob.State.PENDING, PbsJob.State.RUNNING]

    def kill_job(self, name):
        '''
        Kill a job which has the specified name.

        Parameters
        ----------
        name : str

        Returns
        -------
        killed : bool
        '''
        raise NotImplementedError('This method should be implemented by subclasses')

    def get_jobs(self, use_cache=True):
        '''
        Retrieve all the jobs that are currently managed by job scheduler.

        It calls `scontrol show job` for Slurm or `qstat -f -u` for Torque to get the list of jobs.
        If some jobs have finished long time ago (depends on the setting of job scheduler on the machine),
        they may disappear from the list output by job scheduler.

        In order not to apply too much pressure to the job scheduler, cache can be used to stores the jobs.
        If `use_cache` set to True and the cache is not expired, this method will return the cached results without calling `scontrol` or `qstat`.
        The lifetime of cache is determined by attribute :attr:`cached_jobs_expire` in seconds.

        Parameters
        ----------
        use_cache : bool
            Whether the cached job list be used

        Returns
        -------
        jobs : list of PbsJob
        '''
        if not use_cache or \
                self._cache_last_update is None \
                or (datetime.datetime.now() - self._cache_last_update).total_seconds() >= self.cached_jobs_expire:
            logger.info('Updating job information...')
            self._cached_jobs = []
            jobs = []
            for i in [3, 2, 1]:
                # in case _get_jobs() raise Exception
                try:
                    jobs = self._get_jobs()
                except Exception as e:
                    logger.warning(repr(e))
                if jobs:
                    break
                # in case _get_jobs() failed, retry after 1s
                if i > 1:
                    time.sleep(1)
            for job in reversed(jobs):  # reverse the job list, in case jobs with same name
                if job not in self._cached_jobs:
                    self._cached_jobs.append(job)

            self._cached_jobs.reverse()  # reverse the job list
            self._cache_last_update = datetime.datetime.now()

        return self._cached_jobs

    def _get_jobs(self):
        '''
        Retrieve all the jobs that are currently managed by job scheduler.

        It calls `scontrol show job` for Slurm or `qstat -f -u` for Torque to get the list of jobs.
        If some jobs have finished long max_running_hour ago (depends on the setting of job scheduler on the machine),
        they may disappear from the list outputted by job scheduler.

        Returns
        -------
        jobs : list of PbsJob
        '''
        raise NotImplementedError('This method should be implemented by subclasses')
